#include "xmldom.h"

#include <assert.h>
#include <stdio.h>

#include <expat.h>

namespace xml {

void elstart(void *userData, const XML_Char *name, const XML_Char **atts)
{
    Parser::Context *pc = static_cast<Parser::Context *>(userData);
    Element *element = new Element();
    element->name = name;
    for (const XML_Char **a = atts; a[0] != NULL && a[1] != NULL; a += 2) {
        element->attributes[a[0]] = a[1];
    }
    if (pc->stack.size() == 0) {
        pc->document->root = owned_ptr<Element>(element);
    } else {
        pc->stack.back()->children.push_back(element);
    }
    pc->stack.push_back(element);
}

static void elend(void *userData, const XML_Char *name)
{
    Parser::Context *pc = static_cast<Parser::Context *>(userData);
    assert(pc->stack.size() > 0);
    assert(pc->stack.back()->name.compare(name) == 0);
    pc->stack.pop_back();
}

static void chardata(void *userData, const XML_Char *s, int len)
{
    Parser::Context *pc = static_cast<Parser::Context *>(userData);
    assert(pc->stack.size() > 0);
    pc->stack.back()->children.push_back(new Text(std::string(s, len)));
}

Element::~Element()
{
    for (std::vector<const Node *>::const_iterator c = children.begin(); c != children.end(); c++) {
        delete *c;
    }
}

const Element *Element::getFirstChild(const char name[]) const
{
    for (std::vector<const Node *>::const_iterator c = children.begin(); c != children.end(); c++) {
        if ((*c)->type == Node::typeElement && static_cast<const Element *>(*c)->name.compare(name) == 0) {
            return static_cast<const Element *>(*c);
        }
    }
    return NULL;
}

std::vector<const Element *> Element::getElements(const char name[]) const
{
    std::vector<const Element *> r;
    for (std::vector<const Node *>::const_iterator c = children.begin(); c != children.end(); c++) {
        if ((*c)->type == Node::typeElement && static_cast<const Element *>(*c)->name.compare(name) == 0) {
            r.push_back(static_cast<const Element *>(*c));
        }
    }
    return r;
}

std::string Element::getAttribute(const char name[]) const
{
    std::map<std::string, std::string>::const_iterator a = attributes.find(name);
    if (a == attributes.end()) {
        return "";
    }
    return (*a).second;
}

std::string Element::getText() const
{
    std::string r;
    for (std::vector<const Node *>::const_iterator i = children.begin(); i != children.end(); i++) {
        if ((*i)->type == Node::typeText) {
            r += static_cast<const Text *>(*i)->text;
        }
    }
    return r;
}

Parser::Parser()
 : parser(XML_ParserCreate(NULL)), context(&doc)
{
    XML_SetElementHandler(parser, elstart, elend);
    XML_SetCharacterDataHandler(parser, chardata);
    XML_SetUserData(parser, &context);
}

Parser::~Parser()
{
    XML_ParserFree(parser);
}

const Document &Parser::getDocument()
{
    if (XML_Parse(parser, NULL, 0, 1) == 0) {
        doc.clear();
    }
    return doc;
}

bool Parser::parse(const char buf[], int len)
{
    if (len < 0) {
        len = strlen(buf);
    }
    return XML_Parse(parser, buf, len, 0) != 0;
}

int Parser::getErrorCode()
{
    return XML_GetErrorCode(parser);
} 

std::string Parser::getErrorString()
{
    std::string error(XML_ErrorString(getErrorCode()));
    return error;
}

int Parser::getErrorLineNumber()
{
    return XML_GetCurrentLineNumber(parser);
}

Document parseDocumentFile(const char fn[])
{
    Parser parser;
    FILE *f = fopen(fn, "r");
    if (f != NULL) {
        for (;;) {
            char buf[1024];
            int n = fread(buf, 1, sizeof(buf), f);
            if (n <= 0) {
                break;
            }
            if (!parser.parse(buf, n)) {
                fclose(f);
                return Document();
            }
        }
        fclose(f);
    }
    return parser.getDocument();
}

Document parseDocumentText(const char xml[])
{
    Parser parser;
    if (!parser.parse(xml)) {
        return Document();
    }
    return parser.getDocument();
}

} // namespace xml

