#ifndef XMLDOM_H_
#define XMLDOM_H_

#pragma warning(disable: 4786) // debug info too long

#include <deque>
#include <map>
#include <string>
#include <vector>

#include "owned_ptr.h"

namespace xml {

class Node {
public:
    enum NodeType {typeElement, typeText};
    Node(NodeType t): type(t) {}
    virtual ~Node() {}
    const NodeType type;
    virtual std::string getText() const = 0;
private:
    Node(const Node &);
    const Node &operator =(const Node &);
};

class Element: public Node {
public:
    Element(): Node(typeElement) {}
    virtual ~Element();
    std::string name;
    std::map<std::string, std::string> attributes;
    std::vector<const Node *> children;
    const Element *getFirstChild(const char name[]) const;
    std::vector<const Element *> getElements(const char name[]) const;
    std::string getAttribute(const char attr[]) const;
    virtual std::string getText() const;
private:
    Element(const Element &);
    Element &operator =(const Element &);
};

class Text: public Node {
public:
    Text(const std::string &t): Node(typeText), text(t) {}
    virtual std::string getText() const { return text; }
    std::string text;
private:
    Text(const Text &);
    Text &operator =(const Text &);
};

class Document {
public:
    Document() {}
    void clear() { root = owned_ptr<Element>(); }
    const Element *getRootElement() const { return root.get(); }
private:
    owned_ptr<Element> root;
    friend void elstart(void *, const char *, const char **);
};

class Parser {
public:
    Parser();
    ~Parser();
    bool parse(const char buf[], int len = -1);
    const Document &getDocument();
    int getErrorCode();
    std::string getErrorString();
    int getErrorLineNumber();


    class Context {
    public:
        Context(Document *d): document(d) {}
        Document *document;
        std::deque<Element *> stack;
    };

private:
    void *parser;
    Document doc;
    Context context;
private:
    Parser(const Parser &);
    Parser &operator =(const Parser &);
};

Document parseDocumentFile(const char fn[]);
Document parseDocumentText(const char xml[]);

} // namespace xml

#endif /* XMLDOM_H_ */
